#!/usr/bin/env python

"""
Preparation of run-time attribute lookup tables.

Copyright (C) 2007, 2008 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from micropython.common import *
try:
    set
except NameError:
    from sets import Set as set

class TableError(ProcessingError): pass

class List:

    """
    A displaced list containing attribute details and an index mapping names to
    offsets in the list. This is the optimised form of the table initially
    constructed to record object attributes.
    """

    def __init__(self, names):
        self.names = names
        self.displaced = []
        self.offsets = {}
        self.offsets_used = set()

        self.names_index = {}
        for i, name in enumerate(self.names):
            self.names_index[name] = i

    def __len__(self):
        return len(self.displaced)

    # Simulation methods.

    def access(self, objname, attrname):
        obj_offset = self.offsets.get(objname)
        attr_index = self.names_index.get(attrname)
        if obj_offset is None or attr_index is None:
            return None

        element = self.displaced[obj_offset + attr_index]
        if element is not None:
            offset, details = element
            if offset == obj_offset:
                return details

        return None

    def dir(self, objname):
        attributes = {}
        for name in self.names:
            attr = self.access(objname, name)
            if attr is not None:
                attributes[name] = attr
        return attributes

    # Update methods.

    def add(self, objname, attributes):

        "Add an entry for 'objname' with the given 'attributes'."

        len_displaced = len(self.displaced)
        len_attributes = len(attributes)

        # Try to fit the row into the list, starting from the beginning and
        # skipping places where an existing row has been fitted.

        for offset in xrange(0, len_displaced):
            if offset in self.offsets_used:
                continue
            if self._fit_row(offset, attributes, len_displaced):
                self.offsets_used.add(offset)
                self.offsets[objname] = offset
                self._add_row(offset, attributes, len_displaced)
                break

        # Where a row could not be fitted within the list, add it to the
        # end.

        else:
            self.offsets_used.add(len_displaced)
            self.offsets[objname] = len_displaced
            self._add_row(len_displaced, attributes, len_displaced)

    def _fit_row(self, offset, attributes, len_displaced):

        """
        Fit, at the given 'offset', the row of 'attributes' in the displaced
        list having length 'len_displaced'.
        Return a true value if this succeeded.
        """

        for i, attr in enumerate(attributes):
            if i + offset >= len_displaced:
                break
            element = self.displaced[i + offset]
            if attr is not None and element is not None:
                return 0
        return 1

    def _add_row(self, offset, attributes, len_displaced):

        """
        Add, at the given 'offset', the row of 'attributes' in the displaced
        list having length 'len_displaced'.
        """

        # Extend the list if necessary.

        for i in xrange(0, offset + len(attributes) - len_displaced):
            self.displaced.append(None)

        # Record the offset and attribute details in the list.

        for i, attr in enumerate(attributes):
            if attr is not None:
                self.displaced[offset+i] = offset, attr

class Table:

    "A lookup table."

    TableError = TableError

    def __init__(self):
        self.attributes = set()
        self.table = {}
        self.objnames = []
        self.names = []
        self.displaced_list = None

    def add(self, objname, attributes):

        "For the given 'objname' add the given 'attributes' to the table."

        self.table[objname] = attributes
        for name, origin in attributes.items():
            self.attributes.add(name)

        if self.displaced_list is not None:
            self.displaced_list.add(objname, self.matrix_row(attributes))

    def object_names(self):

        "Return the object names used in the table."

        self.objnames = self.objnames or list(self.table.keys())
        return self.objnames

    def attribute_names(self):

        "Return the attribute names used in the table."

        self.names = self.names or list(self.attributes)
        return self.names

    def get_code(self, name):

        "Return the code of the given 'name'."

        try:
            return self.object_names().index(name)
        except ValueError:
            raise TableError, "Name %r is not registered as an object in the table." % name

    def get_index(self, name):

        "Return the index of the given 'name'."

        try:
            return self.attribute_names().index(name)
        except ValueError:
            raise TableError, "Name %r is not registered as an attribute in the table." % name

    def as_matrix(self):

        """
        Return the table as a matrix mapping object names to lists of attributes
        arranged in the order of the recorded attribute names.
        """

        matrix = {}
        for objname, attributes in self.table.items():
            matrix[objname] = self.matrix_row(attributes)
        return matrix

    def matrix_row(self, attributes):

        "Return a matrix row for the given 'attributes'."

        row = []
        for name in self.attribute_names():
            row.append(attributes.get(name))
        return row

    def as_list(self):

        """
        Return a displaced list object encoding the table in a more compact form
        than that provided by the matrix representation. This list will be
        updated with new entries if added to this object using the 'add' method.
        """

        if self.displaced_list is None:
            self.displaced_list = List(self.attribute_names())

            # Visit each row of the matrix.

            for objname, attributes in self.as_matrix().items():
                self.displaced_list.add(objname, attributes)

        return self.displaced_list

# vim: tabstop=4 expandtab shiftwidth=4
